import pytest

from apppy.fs import FileSystem, FileUrl
from apppy.sb.fs import SupabaseFileSystem
from apppy.sb.fs_fixtures import (  # noqa: F401
    sb_file_write,
    sb_file_write_encrypted,
    sb_fs,
    sb_fs_encrypted,
)


def test_load_sb_proxyfs_by_protocol(sb_fs: FileSystem):  # noqa: F811
    _, proxyfs = sb_fs.load_proxyfs_by_protocol("sb")
    assert proxyfs is not None
    assert isinstance(proxyfs, SupabaseFileSystem) is True


async def test_sb_write_bytes_unencrypted(
    sb_fs: FileSystem,  # noqa: F811
):
    file_url = sb_fs.new_file_url_external(
        protocol="sb",
        external_id=None,
        partition="fs_sb_test",
        directory="test_sb_write_bytes_unencrypted",
        file_name="file.txt",
    )

    assert file_url.filesystem_external_id is not None
    assert file_url.is_encrypted is False

    await sb_fs.write_bytes(
        file_url, bytes("The quick brown fox jumped over the lazy dogs.", "utf-8")
    )
    read_bytes = sb_fs.read_bytes(file_url)
    assert read_bytes == bytes("The quick brown fox jumped over the lazy dogs.", "utf-8")


async def test_sb_write_bytes_encrypted(
    sb_fs_encrypted: FileSystem,  # noqa: F811
):
    file_url = sb_fs_encrypted.new_file_url_external(
        protocol="sb",
        external_id=None,
        partition="fs_sb_test",
        directory="test_sb_write_bytes_encrypted",
        file_name="file.txt",
    )

    assert file_url.filesystem_external_id is not None
    assert file_url.is_encrypted is True

    await sb_fs_encrypted.write_bytes(
        file_url, bytes("The quick brown fox jumped over the lazy dogs.", "utf-8")
    )
    read_bytes = sb_fs_encrypted.read_bytes(file_url)
    assert read_bytes == bytes("The quick brown fox jumped over the lazy dogs.", "utf-8")


async def test_sb_write_text_unencrypted(
    sb_fs: FileSystem,  # noqa: F811
):
    file_url = sb_fs.new_file_url_external(
        protocol="sb",
        external_id=None,
        partition="fs_sb_test",
        directory="test_sb_write_text_unencrypted",
        file_name="file.txt",
    )

    assert file_url.filesystem_external_id is not None
    assert file_url.is_encrypted is False

    await sb_fs.write_text(file_url, "The quick brown fox jumped over the lazy dogs.")
    read_bytes = sb_fs.read_bytes(file_url)
    assert read_bytes == bytes("The quick brown fox jumped over the lazy dogs.", "utf-8")


async def test_sb_write_text_encrypted(
    sb_fs_encrypted: FileSystem,  # noqa: F811
):
    file_url = sb_fs_encrypted.new_file_url_external(
        protocol="sb",
        external_id=None,
        partition="fs_sb_test",
        directory="test_sb_write_text_encrypted",
        file_name="file.txt",
    )

    assert file_url.filesystem_external_id is not None
    assert file_url.is_encrypted is True

    await sb_fs_encrypted.write_text(file_url, "The quick brown fox jumped over the lazy dogs.")
    read_bytes = sb_fs_encrypted.read_bytes(file_url)
    assert read_bytes == bytes("The quick brown fox jumped over the lazy dogs.", "utf-8")


@pytest.mark.parametrize(
    ("algo", "expected_checksum"),
    [
        (
            "md5",
            "5c9f966da28ab24ca7796006a6259494",
        ),
        (
            "sha256",
            "c9c85caa5a93aad2bfcc91b9a02d4185a0f0348aac049e650bd0f4dea10a7393",
        ),
    ],
)
async def test_sb_checksum_file_unencrypted(
    sb_fs: FileSystem,  # noqa: F811
    sb_file_write,  # noqa: F811
    algo,
    expected_checksum,
):
    file_url: FileUrl = await sb_file_write(
        partition="fs_sb_test", directory="test_sb_checksum_file"
    )
    assert file_url.is_encrypted is False

    checksum = sb_fs.checksum_file(file_url, algo=algo)
    assert checksum == expected_checksum


@pytest.mark.parametrize(
    ("algo", "expected_checksum"),
    [
        (
            "md5",
            "5c9f966da28ab24ca7796006a6259494",
        ),
        (
            "sha256",
            "c9c85caa5a93aad2bfcc91b9a02d4185a0f0348aac049e650bd0f4dea10a7393",
        ),
    ],
)
async def test_sb_checksum_file_encrypted(
    sb_fs_encrypted: FileSystem,  # noqa: F811
    sb_file_write_encrypted,  # noqa: F811
    algo,
    expected_checksum,
):
    file_url: FileUrl = await sb_file_write_encrypted(
        partition="fs_sb_test", directory="test_sb_checksum_file"
    )
    assert file_url.is_encrypted is True

    checksum = sb_fs_encrypted.checksum_file(file_url, algo=algo)
    assert checksum == expected_checksum



from apppy.fastql.annotation import fastql_type_error
from apppy.fastql.errors import GraphQLServerError


@fastql_type_error
class MissingUserIdentityError(GraphQLServerError):
    """Error raised when a user's identity is missing"""

    def __init__(self) -> None:
        super().__init__("missing_user_identity")


@fastql_type_error
class UserIdentityServerError(GraphQLServerError):
    """Error raised when a user identity flow encountered an error on the server side"""

    def __init__(self, code: str) -> None:
        super().__init__(code)



import datetime
import io
import time
import uuid
from pathlib import Path
from typing import Any

import tusclient
import tusclient.client
import tusclient.uploader
from fastapi_lifespan_manager import LifespanManager
from fsspec import register_implementation
from fsspec.spec import AbstractBufferedFile, AbstractFileSystem
from pydantic import Field
from storage3.types import CreateOrUpdateBucketOptions
from supabase.client import (
    AsyncClient as NativeSupabaseAsyncClient,
)
from supabase.client import (
    AsyncSupabaseStorageClient as NativeSupabaseStorageAsyncClient,
)
from supabase.client import (
    Client as NativeSupabaseClient,
)
from supabase.client import (
    SupabaseStorageClient as NativeSupabaseStorageClient,
)
from supabase.client import (
    create_async_client,
    create_client,
)

from apppy.env import Env, EnvSettings
from apppy.fs import (
    FileSystem,
    FileSystemBucket,
    FileSystemPermission,
    FileUrl,
    GenericFileUrl,
    ProxyFileSystem,
)
from apppy.fs.errors import FileSystemInvalidProtocolError, MalformedFileUrlError
from apppy.generic.encrypt import BytesEncrypter
from apppy.logger import WithLogger


class SupabaseFileUrl(GenericFileUrl):
    def __init__(
        self,
        _filesystem_protocol: str,
        _filesystem_bucket: FileSystemBucket,
        _filesystem_external_id: str | None,
        _partition: str,
        _directory: str | None,
        _file_name: str | None,
    ) -> None:
        super().__init__(
            _filesystem_protocol=_filesystem_protocol,
            _filesystem_bucket=_filesystem_bucket,
            _filesystem_external_id=_filesystem_external_id,
            _partition=_partition,
            _directory=_directory,
            _file_name=_file_name,
        )
        str_instance = self.as_str_internal()

        # Validation
        if _filesystem_protocol != "enc://sb" and _filesystem_protocol != "sb":
            raise FileSystemInvalidProtocolError(protocol=_filesystem_protocol)
        # The id and file name must either both be None or not None
        if _filesystem_external_id is None and _file_name is not None:
            raise MalformedFileUrlError(
                url=str_instance, code="supabase_file_url_file_name_without_external_id"
            )
        elif _filesystem_external_id is not None and _file_name is None:
            raise MalformedFileUrlError(
                url=str_instance, code="supabase_file_url_external_id_without_file_name"
            )

        self._key_prefix = str_instance[
            len(f"{self.filesystem_protocol}://{_filesystem_bucket.value}") + 1 :
        ]
        self._key_prefix_parent = str(Path(self.key_prefix).parent)

    @property
    def key_prefix(self) -> str:
        return self._key_prefix

    @property
    def key_prefix_parent(self) -> str:
        return self._key_prefix_parent

    @staticmethod
    def split_path(path: str, protocol: str, bucket: FileSystemBucket) -> "SupabaseFileUrl":
        generic_file_url = GenericFileUrl.split_path(path=path, protocol=protocol, bucket=bucket)

        return SupabaseFileUrl(
            _filesystem_protocol=generic_file_url.filesystem_protocol,
            _filesystem_bucket=bucket,
            _filesystem_external_id=generic_file_url.filesystem_external_id,
            _partition=generic_file_url.partition,
            _directory=generic_file_url.directory,
            _file_name=generic_file_url.file_name,
        )

    @staticmethod
    def split_url(url: str, bucket: FileSystemBucket) -> "SupabaseFileUrl":
        url = url.strip()
        protocol = GenericFileUrl._parse_protocol(url, unencrypted=False)
        path = url[len(f"{protocol}://") :]

        return SupabaseFileUrl.split_path(path=path, protocol=protocol, bucket=bucket)


DEFAULT_BLOCK_SIZE = 6 * 1024 * 1024
SYNTHETIC_FILE_SIZE = 2**63 - 1


class SupabaseFile(AbstractBufferedFile):
    def __init__(
        self,
        fs,
        path,
        mode="rb",
        block_size=DEFAULT_BLOCK_SIZE,
        autocommit=True,
        cache_type="readahead",
        cache_options=None,
        size=None,
        **kwargs,
    ):
        super().__init__(
            fs=fs,
            path=path,
            mode=mode,
            block_size=block_size,
            autocommit=autocommit,
            cache_type=cache_type,
            cache_options=cache_options,
            # The AbstractBufferedFile base constructor will attempt to lookup
            # the file size if one is not provided. It's been observed that there's
            # a race condition here because Supabase needs some time to index the
            # new file. So if the caller is not providing a size, we'll attach a
            # synthetic size value to delay the file size lookup.
            size=(size or SYNTHETIC_FILE_SIZE),
            **kwargs,
        )
        self.sb_url: SupabaseFileUrl = SupabaseFileUrl.split_path(
            path=path, protocol=fs.protocol, bucket=kwargs["filesystem_bucket"]
        )

        self._tus_client: tusclient.client.TusClient | None = None
        self._tus_uploader: tusclient.uploader.Uploader | None = None

    @property
    def __tus_uploader(self) -> tusclient.uploader.Uploader:
        if self._tus_uploader is None:
            raise Exception("tus_uploader is uninitialized")

        return self._tus_uploader

    def commit(self):
        """Move from temp to final destination"""
        self._upload_chunk(final=True)

    @property
    def details(self):
        if self._details is None:  # type: ignore[has-type]
            self._details = self._fetch_details_with_retry()
            # See constructor comments about synthetic file size
            # If we encounter this and have a real file size available
            # update the file details to use the real size.
            if self.size == SYNTHETIC_FILE_SIZE:
                real_size = self._details.get("size")
                if isinstance(real_size, int) and real_size >= 0:
                    self.size = real_size
                    # Update any cache value too just in case
                    if hasattr(self, "cache") and hasattr(self.cache, "size"):
                        self.cache.size = real_size

        return self._details

    def discard(self):
        """Throw away temporary file"""

    def _fetch_details_with_retry(self, attempts=8, base_sleep=0.05):
        """
        Attempt to retrieve file details with a linear backoff.

        Supabase takes some time in order to index files so there's a brief
        period in which a file may exist but its information cannot be retrieved.
        """
        last_exc: FileNotFoundError | None = None
        for i in range(attempts):
            try:
                return self.fs.info(self.path, **self.kwargs)
            except FileNotFoundError as e:
                last_exc = e
                time.sleep(base_sleep * (i + 1))
        # If still not visible, surface the error only when metadata is required
        raise last_exc  # type: ignore[invalid-raise]

    def _fetch_range(self, start, end):
        """Get the specified set of bytes from remote"""
        downloaded_bytes = (
            self.fs.storage_client()
            .from_(self.sb_url.filesystem_bucket.value)
            .download(self.sb_url.key_prefix)
        )

        file_like = io.BytesIO(downloaded_bytes)
        size = len(downloaded_bytes)

        # Handle start
        if start is not None:
            if start >= 0:
                file_like.seek(start)
            else:
                file_like.seek(max(0, size + start))
        else:
            file_like.seek(0)

        # Handle end
        if end is not None:
            if end < 0:
                end = size + end
            return file_like.read(end - file_like.tell())

        return file_like.read()

    def _initiate_upload(self):
        """Create remote file/upload"""
        self._tus_client = tusclient.client.TusClient(
            f"{self.fs._settings.api_url}/storage/v1/upload/resumable",
            headers={"Authorization": f"Bearer {self.fs._settings.api_key}", "x-upsert": "true"},
        )
        self._tus_uploader = self._tus_client.uploader(
            file_stream=self.buffer,
            chunk_size=self.blocksize,
            metadata={
                "bucketName": self.sb_url.filesystem_bucket.value,
                "objectName": self.sb_url.key_prefix,
                "contentType": "application/octet-stream",
                "cacheControl": "3600",
            },
        )

    def _upload_chunk(self, final=False):
        """Write one part of a multi-block file upload

        Parameters
        ==========
        final: bool
            This is the last block, so should complete file, if
            self.autocommit is True.
        """
        self.__tus_uploader.upload_chunk()
        if final:
            # There's no apparent close function for these
            # so just set them to None to signal that we
            # should not expect to continue with any more
            # uploading
            self._tus_client = None
            self._tus_uploader = None


class SupabaseFileSystemSettings(EnvSettings):
    # SUPABASE_FS_BUCKET_EXTERNAL
    bucket_external: str = Field()
    # SUPABASE_FS_BUCKET_INTERNAL
    bucket_internal: str = Field()
    # SUPABASE_FS_API_KEY
    api_key: str = Field(exclude=True)
    # SUPABASE_FS_API_URL
    api_url: str = Field()
    # SUPABASE_FS_ENCRYPT_PASSPHRASE
    encrypt_passphrase: str | None = Field(default=None, exclude=True)
    # SUPABASE_FS_ENCRYPT_SALT
    encrypt_salt: str | None = Field(default=None, exclude=True)

    def __init__(self, env: Env) -> None:
        super().__init__(env=env, domain_prefix="SUPABASE_FS")


class NativeSupabaseFileSystem(AbstractFileSystem, WithLogger):
    """
    A fsspec plugin that works with Supabase. In most other cases,
    we are able to leverage plugins that ship with fsspec itself.

    However, in this case, there is no native Supabase plugin so
    we need to write this one.
    """

    protocol = "sb"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        storage_options = kwargs["storage_options"]

        self._settings: SupabaseFileSystemSettings = storage_options["settings"]

        self._fs_bucket_names: list[str] = storage_options["fs_bucket_names"]
        self._native_storage_client_async: NativeSupabaseStorageAsyncClient | None = None
        lifespan: LifespanManager = storage_options["lifespan"]
        lifespan.add(self.__configure_supabase_storage)

    ##### ##### ##### Configuration and Registration ##### ##### #####

    async def __configure_supabase_storage(self):
        native_client_async: NativeSupabaseAsyncClient = await create_async_client(
            supabase_url=self._settings.api_url, supabase_key=self._settings.api_key
        )

        self._native_storage_client_async = native_client_async.storage

        # Ensure that all buckets used by this FileSystem
        # are created
        all_buckets = await self.list_buckets()
        fs_buckets = []  # Buckets used by this FileSystem
        for fs_bucket_name in self._fs_bucket_names:
            fs_bucket = next(
                (bucket for bucket in all_buckets if bucket.name == fs_bucket_name), None
            )
            if fs_bucket is None:
                fs_bucket = await self.create_bucket(
                    id=fs_bucket_name,
                    options=CreateOrUpdateBucketOptions(public=False),
                )

            fs_buckets.append(fs_bucket)

        yield {
            "native_storage_client_async": self._native_storage_client_async,
            "fs_buckets": fs_buckets,
        }

        self._logger.info("Closing Supabase storage clients")
        await self._native_storage_client_async.aclose()

    def storage_client(self) -> NativeSupabaseStorageClient:
        # This allows us to create a client for every file system
        # interation. This is super safe as we'll guarantee that
        # the client is never closed. But might we leak clients here?
        # TODO: Is this OK to do?
        native_client: NativeSupabaseClient = create_client(
            supabase_url=self._settings.api_url,
            supabase_key=self._settings.api_key,
        )
        return native_client.storage

    ##### ##### ##### Bucket Management ##### ##### #####

    @property
    def __native_storage_client_async(self) -> NativeSupabaseStorageAsyncClient:
        if self._native_storage_client_async is None:
            raise Exception("native_storage_client_async is uninitialied")

        return self._native_storage_client_async

    async def create_bucket(
        self, id: str, name: str | None = None, options: CreateOrUpdateBucketOptions | None = None
    ) -> dict[str, str]:
        return await self.__native_storage_client_async.create_bucket(
            id=id, name=name, options=options
        )

    async def get_bucket(self, id: str):
        return await self.__native_storage_client_async.get_bucket(id=id)

    async def list_buckets(self):
        return await self.__native_storage_client_async.list_buckets()

    ##### ##### ##### FileSystem Implementation ##### ##### #####

    @property
    def fsid(self):
        return "supabase"

    def created(self, path):
        """Return the created timestamp of a file as a datetime.datetime"""
        path_info = self.info(path)
        created_dt = datetime.datetime.fromisoformat(path_info["created"])
        return created_dt

    def cp_file(self, path1, path2, **kwargs):
        sb_file_url1: SupabaseFileUrl = SupabaseFileUrl.split_path(
            path=path1, protocol=self.protocol, bucket=kwargs["filesystem_bucket"]
        )
        sb_file_url2: SupabaseFileUrl = SupabaseFileUrl.split_path(
            path=path2, protocol=self.protocol, bucket=kwargs["filesystem_bucket"]
        )
        # Ensure that we're only copying within the same bucket
        # Copying between buckets is not allowed
        assert sb_file_url1.filesystem_bucket.value == sb_file_url2.filesystem_bucket.value

        self.storage_client().from_(sb_file_url1.filesystem_bucket.value).copy(
            from_path=sb_file_url1.key_prefix,
            to_path=sb_file_url2.key_prefix,
        )

    def info(self, path, **kwargs):
        # For Supabase, treat info just as a straight
        # alias for ls
        path = self._strip_protocol(path)
        sb_file_url = SupabaseFileUrl.split_path(
            path=path, protocol=self.protocol, bucket=kwargs["filesystem_bucket"]
        )
        if sb_file_url.is_file is False:
            if sb_file_url.key_prefix is None:  # The path cannot be parsed
                raise FileNotFoundError

            # Supabase does not really have the concept of
            # directories so we'll fake some data here and
            # just return it. Note that this may create upstream
            # consequences. For example, directories will always
            # be seen as existing when the key prefix is in use or not.
            return {
                "name": Path(sb_file_url.key_prefix).name,
                "size": None,
                "type": "directory",
                "url": path,
            }

        resp = self.ls(path, detail=True, **kwargs)
        if len(resp) == 0:
            raise FileNotFoundError

        return resp[0]

    def isdir(self, path, **kwargs):
        path = self._strip_protocol(path)
        sb_file_url = SupabaseFileUrl.split_path(
            path=path, protocol=self.protocol, bucket=kwargs["filesystem_bucket"]
        )

        return sb_file_url.is_directory

    def isfile(self, path, **kwargs):
        path = self._strip_protocol(path)
        sb_file_url = SupabaseFileUrl.split_path(
            path=path, protocol=self.protocol, bucket=kwargs["filesystem_bucket"]
        )

        return sb_file_url.is_file

    def ls(self, path, detail=True, **kwargs):
        """List objects at path.

        This should include subdirectories and files at that location. The
        difference between a file and a directory must be clear when details
        are requested.

        The specific keys, or perhaps a FileInfo class, or similar, is TBD,
        but must be consistent across implementations.
        Must include:

        - full path to the entry (without protocol)
        - size of the entry, in bytes. If the value cannot be determined, will
          be ``None``.
        - type of entry, "file", "directory" or other

        Additional information
        may be present, appropriate to the file-system, e.g., generation,
        checksum, etc.

        May use refresh=True|False to allow use of self._ls_from_cache to
        check for a saved listing and avoid calling the backend. This would be
        common where listing may be expensive.

        Parameters
        ----------
        path: str
        detail: bool
            if True, gives a list of dictionaries, where each is the same as
            the result of ``info(path)``. If False, gives a list of paths
            (str).
        kwargs: may have additional backend-specific options, such as version
            information

        Returns
        -------
        List of strings if detail is False, or list of directory information
        dicts if detail is True.
        """
        path = self._strip_protocol(path)
        sb_file_url = SupabaseFileUrl.split_path(
            path=path, protocol=self.protocol, bucket=kwargs["filesystem_bucket"]
        )
        if sb_file_url.is_file:
            if sb_file_url.key_prefix is None:
                return []  # The path cannot be parsed

            # This is a bit of a hack. It's been observed
            # that we cannot run a list on a file's key prefix
            # So instead, we'll list the parent and filter here
            # for the file name
            resp_parent = (
                self.storage_client()
                .from_(sb_file_url.filesystem_bucket.value)
                .list(sb_file_url.key_prefix_parent)
            )
            resp = [r for r in resp_parent if sb_file_url.key_prefix.endswith(r["name"])]
        else:
            # CASE: Path is a directory
            resp = (
                self.storage_client()
                .from_(sb_file_url.filesystem_bucket.value)
                .list(sb_file_url.key_prefix)
            )

        ls_results = []
        for r in resp:
            ls_result = {
                "name": r["name"],
            }
            if self.isfile(path, **kwargs):
                ls_result["url"] = sb_file_url.as_str_internal()
            else:
                ls_result["url"] = (
                    f"sb://{sb_file_url.filesystem_bucket.value}/{sb_file_url.key_prefix}/{r['name']}"
                )

            if detail:
                ls_result["size"] = (
                    r["metadata"].get("size") if r.get("metadata") is not None else None
                )
                ls_result["type"] = "file"
                ls_result["created"] = r["created_at"]
                ls_result["modified"] = r["created_at"]
                ls_result["islink"] = False
                ls_result["mode"] = "ro"  # Readonly?
                ls_result["uid"] = r["id"]
                ls_result["gid"] = r["id"]

            ls_results.append(ls_result)

        return ls_results

    def lsdir(self, path, recursive=False, maxdepth=None, **kwargs):
        path = self._strip_protocol(path)

        ls_files = []
        current_depth = kwargs.get("current_depth", 0)
        if maxdepth is not None and current_depth > maxdepth:
            return ls_files

        sb_file_url: SupabaseFileUrl = SupabaseFileUrl.split_path(
            path=path, protocol=self.protocol, bucket=kwargs["filesystem_bucket"]
        )

        if sb_file_url.is_file:
            ls_files.append({"name": sb_file_url.key_prefix, "url": sb_file_url.as_str_internal()})
            return ls_files

        ls_results = self.ls(path, detail=True, **kwargs)
        for ls_result in ls_results:
            ls_result_path = self._strip_protocol(ls_result["url"])
            ls_result_url: SupabaseFileUrl = SupabaseFileUrl.split_path(
                path=ls_result_path, protocol=self.protocol, bucket=kwargs["filesystem_bucket"]
            )
            ls_files.append(
                {"name": ls_result_url.key_prefix, "url": ls_result_url.as_str_internal()}
            )
            if ls_result_url.is_directory and recursive is True:
                kwargs["current_depth"] = current_depth + 1
                ls_files.extend(
                    self.lsdir(
                        ls_result_path,
                        recursive=True,
                        maxdepth=maxdepth,
                        **kwargs,
                    )
                )

        return ls_files

    def mkdir(self, path, create_parents=True, **kwargs):
        """Create directory entry at path"""
        # Legitimately skip as Supabase storage
        # does not have directories
        pass

    def makedirs(self, path, exist_ok=False):
        """Recursively make directories"""
        # Legitimately skip as Supabase storage
        # does not have directories
        pass

    def modified(self, path):
        """Return the modified timestamp of a file as a datetime.datetime"""
        path_info = self.info(path)
        modified_dt = datetime.datetime.fromisoformat(path_info["modified"])
        return modified_dt

    def rmdir(self, path):
        """Remove a directory, if empty"""
        # Legitimately skip as Supabase storage
        # does not have directories
        pass

    def sign(self, path, expiration=(1 * 60 * 60), **kwargs):
        """
        Create a signed URL representing the given path

        Parameters
        ----------
        path : str
             The path on the filesystem
        expiration : int
            Number of seconds to enable the URL for (if supported)

        Returns
        -------
        URL : str
            The signed URL
        """
        path = self._strip_protocol(path)
        sb_file_url: SupabaseFileUrl = SupabaseFileUrl.split_path(
            path=path, protocol=self.protocol, bucket=kwargs["filesystem_bucket"]
        )
        resp = (
            self.storage_client()
            .from_(sb_file_url.filesystem_bucket.value)
            .create_signed_url(
                path=sb_file_url.key_prefix,
                expires_in=expiration,
            )
        )

        return resp["signedURL"]

    def _open(
        self,
        path,
        mode="rb",
        block_size=None,
        autocommit=True,
        cache_options=None,
        **kwargs,
    ):
        """Return raw bytes-mode file-like from the file-system"""
        return SupabaseFile(
            fs=self,
            path=path,
            mode=mode,
            # block_size=block_size,
            autocommit=autocommit,
            cache_options=cache_options,
            **kwargs,
        )

    def _rm(self, path, recursive=False, maxdepth=None, **kwargs):
        path = self._strip_protocol(path)
        sb_file_url: SupabaseFileUrl = SupabaseFileUrl.split_path(
            path=path, protocol=self.protocol, bucket=kwargs["filesystem_bucket"]
        )

        ls_dir_results = self.lsdir(path, recursive, maxdepth, **kwargs)
        key_prefixes = []
        for ls_result in ls_dir_results:
            ls_result_path = self._strip_protocol(ls_result["url"])
            ls_result_file_url = SupabaseFileUrl.split_path(
                path=ls_result_path, protocol=self.protocol, bucket=kwargs["filesystem_bucket"]
            )
            key_prefixes.append(ls_result_file_url.key_prefix)

        if len(key_prefixes) > 0:
            self.storage_client().from_(sb_file_url.filesystem_bucket.value).remove(key_prefixes)

    @classmethod
    def _strip_protocol(cls, path):
        # CASE: The protocol has already been stripped
        if path.find("://") == -1:
            return path

        protocol = FileUrl._parse_protocol(path, unencrypted=False)
        return path[len(protocol) + 3 :]


class SupabaseFileSystem(ProxyFileSystem, WithLogger):
    def __init__(
        self,
        env: Env,
        settings: SupabaseFileSystemSettings,
        lifespan: LifespanManager,
        fs: FileSystem,
    ) -> None:
        self._settings: SupabaseFileSystemSettings = settings

        self._bytes_encrypter: BytesEncrypter | None = None
        if (
            settings.encrypt_passphrase is not None
            and len(settings.encrypt_passphrase) > 0
            and settings.encrypt_salt is not None
            and len(settings.encrypt_salt) > 0
        ):
            self._bytes_encrypter = BytesEncrypter(
                settings.encrypt_passphrase, settings.encrypt_salt
            )

        self.__configure_nativefs(env, settings, lifespan, fs)

    def __configure_nativefs(
        self,
        env: Env,
        settings: SupabaseFileSystemSettings,
        lifespan: LifespanManager,
        fs: FileSystem,
    ) -> None:
        # Use generic test buckets for all tests
        bucket_external_value = (
            f"{settings.bucket_external}-test"
            if env.is_test or env.is_ci
            else settings.bucket_external
        )
        self._bucket_external = FileSystemBucket(
            bucket_type="external",
            value=bucket_external_value,
        )
        bucket_internal_value = (
            f"{settings.bucket_internal}-test"
            if env.is_test or env.is_ci
            else settings.bucket_internal
        )
        self._bucket_internal = FileSystemBucket(
            bucket_type="internal",
            value=bucket_internal_value,
        )
        fs.register_proxyfs(self, "sb")

        fs_bucket_names: list[str] = [
            self._bucket_external.value,
            self._bucket_internal.value,
        ]

        nativefs = NativeSupabaseFileSystem(
            storage_options={
                "fs_bucket_names": fs_bucket_names,
                "settings": settings,
                "lifespan": lifespan,
            }
        )
        self._nativefs: AbstractFileSystem = nativefs
        fs.register_nativefs(nativefs, "sb")

        if self.is_encrypted:
            # In the encrypted case, we'll need to also register the file
            # system with fsspec itself so that the encrypted filesystem
            # can instantiate it independently
            register_implementation("sb", NativeSupabaseFileSystem, clobber=True)

    def convert_file_url(self, file_url: FileUrl) -> FileUrl:
        if isinstance(file_url, SupabaseFileUrl) and self.is_encrypted == file_url.is_encrypted:
            return file_url

        if self.is_encrypted is True and file_url.is_encrypted is False:
            # Add the encyption protocol if we need to
            filesystem_protocol = f"enc://{file_url.filesystem_protocol}"
        else:
            filesystem_protocol = file_url.filesystem_protocol

        # Always ensure that we always have a unique id associated with the file_url.
        # Unfortunately, it's not possible to get the actual unique identiter out of
        # Supabase storage so we'll just make one up here and pass it along
        filesystem_external_id: str | None = None
        if file_url.file_name is not None:
            filesystem_external_id = (
                str(uuid.uuid4())
                if file_url.filesystem_external_id is None
                else file_url.filesystem_external_id
            )
        return SupabaseFileUrl(
            _filesystem_protocol=filesystem_protocol,
            _filesystem_bucket=(
                self._bucket_external
                if file_url.filesystem_bucket.is_external
                else self._bucket_internal
            ),
            _filesystem_external_id=filesystem_external_id,
            _partition=file_url.partition,
            _directory=file_url.directory,
            _file_name=file_url.file_name,
        )

    def file_url_kwargs(self, file_url: FileUrl) -> dict[str, Any]:
        # The NativeSupabaseFileSystem relies on us providing
        # the filesystem_bucket as an extra parameter
        return {"filesystem_bucket": file_url.filesystem_bucket}

    def parse_file_url(self, url: str) -> SupabaseFileUrl:
        if url.find(self._bucket_internal.value) > -1:
            return SupabaseFileUrl.split_url(url, self._bucket_internal)

        return SupabaseFileUrl.split_url(url, self._bucket_external)

    def rm(self, url: str, recursive=False, maxdepth=None, **kwargs) -> None:
        self.native.rm(url, recursive=recursive, maxdepth=maxdepth, **kwargs)

    @property
    def encryption(self) -> BytesEncrypter | None:
        return self._bytes_encrypter

    @property
    def name(self) -> str:
        return "Supabase"

    @property
    def native(self) -> AbstractFileSystem:
        return self._nativefs

    @property
    def permissions(self) -> list[FileSystemPermission]:
        return [
            FileSystemPermission.PRIVATE_INTERNAL,
            FileSystemPermission.READWRITE,
        ]



import psycopg

from apppy.db.migrations import Migrations
from apppy.db.postgres import PostgresClient


class SupabaseMigrations(Migrations):
    def __init__(
        self,
        postgres: PostgresClient,
    ):
        self._postgres = postgres

    async def head(self) -> str | None:
        latest_migration_query = """
        SELECT version, name
        FROM supabase_migrations.schema_migrations
        ORDER BY version DESC
        LIMIT 1
        """
        try:
            result_set = await self._postgres.db_query_async(latest_migration_query)
            if result_set is None or len(result_set) != 1:
                return None

            return f"{result_set[0]['version']}_{result_set[0]['name']}"
        except psycopg.errors.UndefinedTable:  # type: ignore[unresolved-attribute]
            # It's been observed that supabase will create the schema_migrations
            # table lazily whenever the first migration file is run. So it's valid
            # that the table may not exist yet.
            return None



import json
import logging
import traceback
from dataclasses import dataclass
from typing import Any

import requests
from fastapi_lifespan_manager import LifespanManager
from pydantic import Field

from apppy.env import Env, EnvSettings
from apppy.logger import WithLogger
from apppy.logger.parser import LogRecordParser
from apppy.queues import Queue, QueueHandler


@dataclass
class LogflareRecord:
    """
    Dataclass which helps to translate between the native
    logging.LogRecord and the Logflare schema
    """

    fileline: str
    level: str
    message: str
    logger: str
    extra: dict[str, Any] | None = None
    state: dict[str, Any] | None = None
    traceback: list[str] | None = None

    def to_logflare_json(self) -> str:
        logflare_dict: dict[str, Any] = {
            "message": self.message,
            "metadata": {
                "extra": self.extra,
                "fileline": self.fileline,
                "level": self.level,
                "logger": self.logger,
                "traceback": self.traceback,
            },
        }

        logflare_metadata = logflare_dict["metadata"]
        if self.state is not None:
            for key, value in self.state.items():
                logflare_metadata[key] = value  # type: ignore[index]

        return json.dumps(logflare_dict)

    @classmethod
    def from_log_record(cls, log_record: logging.LogRecord):
        logflare_record = cls(
            fileline=f"{log_record.pathname}/{log_record.filename}:{log_record.lineno}",
            level=log_record.levelname,
            logger=log_record.name,
            message=log_record.getMessage(),
        )

        state_info = LogRecordParser.get_global().parse_state_info(log_record)
        if state_info:
            logflare_record.state = state_info

        extra_info = LogRecordParser.get_global().parse_extra_info(log_record)
        if extra_info:
            logflare_record.extra = extra_info

        if log_record.levelno > logging.INFO and log_record.exc_info:
            exc_type, exc_value, exc_tb = log_record.exc_info
            traceback_flat = "".join(traceback.format_exception(exc_type, exc_value, exc_tb))
            logflare_record.traceback = traceback_flat.splitlines()

        return logflare_record


class LogflareClientSettings(EnvSettings):
    # LOGFLARE_LOGGING_API_KEY
    api_key: str = Field(exclude=True)
    # LOGFLARE_LOGGING_API_URL
    api_url: str = Field()
    # LOGFLARE_LOGGING_API_TIMEOUT
    api_timeout: int = Field(default=2)
    # LOGFLARE_LOGGING_QUEUE_MAX_SIZE
    queue_max_size: int = Field(default=1000)
    # LOGFLARE_LOGGING_QUEUE_SHUTDOWN_TIMEOUT"
    queue_shutdown_timeout: int = Field(default=2)

    def __init__(self, env: Env) -> None:
        super().__init__(env=env, domain_prefix="LOGFLARE_LOGGING")


class LogflareClient(WithLogger):
    def __init__(
        self, env: Env, settings: LogflareClientSettings, lifespan: LifespanManager
    ) -> None:
        self._application_logs_source_name: str | None = None
        # For tests, we're not going to attempt to configure
        # a logging source. Only for legit environments (e.g.
        # local, staging, production, etc).
        if not env.is_ci and not env.is_test:
            self._application_logs_source_name = f"application.logs.{env.name}"

        self._headers = {"Content-Type": "application/json"}
        self._settings: LogflareClientSettings = settings
        lifespan.add(self._configure_application_logs_source)

    async def _configure_application_logs_source(self):
        application_logs_source = None
        if self._application_logs_source_name is not None:
            for source in self.list_sources():
                if source["name"] == self._application_logs_source_name:
                    application_logs_source = source
                    break

            # Create the application logs source if we cannot
            # find the name already in Logflare
            if application_logs_source is None:
                application_logs_source = self.create_source(self._application_logs_source_name)

        yield {"application_logs_source": application_logs_source}

    def create_source(self, source_name: str) -> dict[Any, Any]:
        resp = requests.post(
            f"{self._settings.api_url}/api/sources?api_key={self._settings.api_key}",
            headers=self._headers,
            data=json.dumps({"name": source_name}),
            timeout=self._settings.api_timeout,
        )

        return resp.json()

    async def emit_application_log(self, record: LogflareRecord) -> None:
        # In some select cases (e.g. tests), there is no
        # application logs source so we'll just skip it.
        #
        # NOTE: We need to be careful here that this is
        # set correctly for legitimate cases or else logging
        # may silently fail.
        if self._application_logs_source_name is None:
            return None

        await self._emit_log(self._application_logs_source_name, record)

    async def _emit_log(self, source_name: str, record: LogflareRecord) -> None:
        requests.post(
            f"{self._settings.api_url}/api/logs?source_name={source_name}&api_key={self._settings.api_key}",
            headers=self._headers,
            data=record.to_logflare_json(),
            timeout=self._settings.api_timeout,
        )

    def list_sources(self) -> list[dict]:
        resp = requests.get(
            f"{self._settings.api_url}/api/sources?api_key={self._settings.api_key}",
            headers=self._headers,
            timeout=self._settings.api_timeout,
        )

        return resp.json()


class LogflareLoggingHandler(logging.Handler, QueueHandler, WithLogger):
    def __init__(
        self,
        logflare_client: LogflareClient,
        logflare_queue: Queue,
    ) -> None:
        super().__init__(level=logging.root.level)
        self._logflare_client = logflare_client
        self._logflare_queue = logflare_queue
        self._logflare_queue.register_handler(self)

    def addTo(self, logger: logging.Logger) -> None:
        # This is an interesting case. The logflare client
        # uses urllib3 under the covers. If the log level
        # is turned up too high on this particular logger
        # and we've added the logflare handler, we get into
        # an infinite loop:
        #
        # 1. Emit a log line with urllib3
        # 2. The urllib3 library creates debugging logs
        # 3. The urllib3 debugging logs are in turn
        #    emitted with urllib3 (i.e. back to #1)
        #
        # So here, we'll make sure that we never attach the
        # logflare handler to urllib3 logs
        #
        # Also, as a safety measure, we'll skip Logflare specifically
        # for loggers in this module as well as the queue module to
        # preent possible infinite loops.
        if (
            logger.name == "urllib3.connectionpool"
            or logger.name.startswith("application.logger")
            or logger.name.startswith("application.queue")
        ):
            logger.setLevel(logging.INFO)
            return

        logger.addHandler(self)

    def addToAll(self) -> None:
        logger_dict = logging.Logger.manager.loggerDict
        for logger_name in logger_dict:
            logger = logger_dict.get(logger_name)
            if isinstance(logger, logging.Logger):
                self.addTo(logger)

    def emit(self, record: logging.LogRecord) -> None:
        logflare_record = LogflareRecord.from_log_record(record)
        self._logflare_queue.publish(logflare_record)

    async def handle_exception(self, message: LogflareRecord, exception: BaseException):
        self._logger.error(
            "Failed to emit log to Logflare", extra={"record": message}, exc_info=exception
        )

    async def handle_message(self, message: LogflareRecord):
        await self._logflare_client.emit_application_log(message)



import pytest

from apppy.fs import FileSystemBucket, FileUrl
from apppy.sb.fs import SupabaseFileUrl

_supabase_fs_bucket_external_test = FileSystemBucket(
    bucket_type="external", value="fs_supabase_test"
)

_case_dir_only: SupabaseFileUrl = SupabaseFileUrl(
    _filesystem_protocol="sb",
    _filesystem_bucket=_supabase_fs_bucket_external_test,
    _filesystem_external_id=None,
    _partition="partition",
    _directory="dir",
    _file_name=None,
)
_case_dir_with_subdir: SupabaseFileUrl = SupabaseFileUrl(
    _filesystem_protocol="sb",
    _filesystem_bucket=_supabase_fs_bucket_external_test,
    _filesystem_external_id=None,
    _partition="partition",
    _directory="dir/subdir",
    _file_name=None,
)
# Malformed url for the Supabase case
# _case_file_name_only: SupabaseFileUrl = SupabaseFileUrl(
#     _filesystem_protocol="sb",
#     _filesystem_bucket=_supabase_fs_bucket_external_test,
#     _filesystem_external_id=None,
#     _partition="partition",
#     _directory=None,
#     _file_name="f.txt",
# )
# Malformed url for the Supabase case
# _case_file_name_with_dir: SupabaseFileUrl = SupabaseFileUrl(
#     _filesystem_protocol="sb",
#     _filesystem_bucket=_supabase_fs_bucket_external_test,
#     _filesystem_external_id=None,
#     _partition="partition",
#     _directory="dir",
#     _file_name="f.txt",
# )
# Malformed url for the Supabase case
# _case_unique_id_only: SupabaseFileUrl = SupabaseFileUrl(
#     _filesystem_protocol="sb",
#     _filesystem_bucket=_supabase_fs_bucket_external_test,
#     _filesystem_external_id="123-abc",
#     _partition="partition",
#     _directory=None,
#     _file_name=None,
# )
# Malformed url for the Supabase case
# _case_unique_id_with_dir: SupabaseFileUrl = SupabaseFileUrl(
#     _filesystem_protocol="sb",
#     _filesystem_bucket=_supabase_fs_bucket_external_test,
#     _filesystem_external_id="123-abc",
#     _partition="partition",
#     _directory="dir",
#     _file_name=None,
# )
_case_unique_id_with_dir_and_file_name: SupabaseFileUrl = SupabaseFileUrl(
    _filesystem_protocol="sb",
    _filesystem_bucket=_supabase_fs_bucket_external_test,
    _filesystem_external_id="123-abc",
    _partition="partition",
    _directory="dir",
    _file_name="f.txt",
)
_case_unique_id_with_file_name: SupabaseFileUrl = SupabaseFileUrl(
    _filesystem_protocol="sb",
    _filesystem_bucket=_supabase_fs_bucket_external_test,
    _filesystem_external_id="123-abc",
    _partition="partition",
    _directory=None,
    _file_name="f.txt",
)
_case_unique_id_with_dir_and_file_name_encrypted: SupabaseFileUrl = SupabaseFileUrl(
    _filesystem_protocol="enc://sb",
    _filesystem_bucket=_supabase_fs_bucket_external_test,
    _filesystem_external_id="123-abc",
    _partition="partition",
    _directory="dir",
    _file_name="f.txt",
)


@pytest.mark.parametrize(
    "file_url, expected_str",
    [
        (_case_dir_only, "sb://external/partition/dir"),
        (_case_dir_with_subdir, "sb://external/partition/dir/subdir"),
        (_case_unique_id_with_dir_and_file_name, "sb://external/partition/dir/@123-abc$f.txt"),
        (_case_unique_id_with_file_name, "sb://external/partition/@123-abc$f.txt"),
        (
            _case_unique_id_with_dir_and_file_name_encrypted,
            "enc://sb://external/partition/dir/@123-abc$f.txt",
        ),
    ],
)
def test_supabase_file_url_str(file_url: FileUrl, expected_str: str):
    assert str(file_url) == expected_str


@pytest.mark.parametrize(
    "file_url, expected_str",
    [
        (_case_dir_only, "sb://fs_supabase_test/partition/dir"),
        (_case_dir_with_subdir, "sb://fs_supabase_test/partition/dir/subdir"),
        (
            _case_unique_id_with_dir_and_file_name,
            "sb://fs_supabase_test/partition/dir/@123-abc$f.txt",
        ),
        (_case_unique_id_with_file_name, "sb://fs_supabase_test/partition/@123-abc$f.txt"),
        (
            _case_unique_id_with_dir_and_file_name_encrypted,
            "sb://fs_supabase_test/partition/dir/@123-abc$f.txt",
        ),
    ],
)
def test_supabase_file_url_str_internal(file_url: FileUrl, expected_str: str):
    assert file_url.as_str_internal() == expected_str


@pytest.mark.parametrize(
    "path, expected_file_url",
    [
        ("external/partition/dir", _case_dir_only),
        ("external/partition/dir/subdir", _case_dir_with_subdir),
        ("external/partition/dir/@123-abc$f.txt", _case_unique_id_with_dir_and_file_name),
        ("external/partition/@123-abc$f.txt", _case_unique_id_with_file_name),
    ],
)
def test_supabase_file_url_split_path(path: str, expected_file_url: FileUrl):
    file_url = SupabaseFileUrl.split_path(
        path, protocol="sb", bucket=_supabase_fs_bucket_external_test
    )
    assert file_url == expected_file_url


@pytest.mark.parametrize(
    "path, expected_file_url",
    [
        ("fs_supabase_test/partition/dir", _case_dir_only),
        ("fs_supabase_test/partition/dir/subdir", _case_dir_with_subdir),
        ("fs_supabase_test/partition/dir/@123-abc$f.txt", _case_unique_id_with_dir_and_file_name),
        ("fs_supabase_test/partition/@123-abc$f.txt", _case_unique_id_with_file_name),
    ],
)
def test_supabase_file_url_split_path_unobfuscated(path: str, expected_file_url: FileUrl):
    file_url = SupabaseFileUrl.split_path(
        path, protocol="sb", bucket=_supabase_fs_bucket_external_test
    )
    assert file_url == expected_file_url


@pytest.mark.parametrize(
    "url, expected_file_url",
    [
        ("sb://external/partition/dir", _case_dir_only),
        ("sb://external/partition/dir/subdir", _case_dir_with_subdir),
        ("sb://external/partition/dir/@123-abc$f.txt", _case_unique_id_with_dir_and_file_name),
        ("sb://external/partition/@123-abc$f.txt", _case_unique_id_with_file_name),
        (
            "enc://sb://external/partition/dir/@123-abc$f.txt",
            _case_unique_id_with_dir_and_file_name_encrypted,
        ),
    ],
)
def test_supabase_file_url_split_url(url: str, expected_file_url: FileUrl):
    file_url = SupabaseFileUrl.split_url(url, bucket=_supabase_fs_bucket_external_test)
    assert file_url == expected_file_url


@pytest.mark.parametrize(
    "file_url, expected_key_prefix",
    [
        (_case_dir_only, "partition/dir"),
        (_case_dir_with_subdir, "partition/dir/subdir"),
        (_case_unique_id_with_dir_and_file_name, "partition/dir/@123-abc$f.txt"),
        (_case_unique_id_with_file_name, "partition/@123-abc$f.txt"),
    ],
)
def test_supabase_file_url_key_prefix(file_url: SupabaseFileUrl, expected_key_prefix: str):
    assert file_url.key_prefix == expected_key_prefix


@pytest.mark.parametrize(
    "file_url, expected_key_prefix_parent",
    [
        (_case_dir_only, "partition"),
        (_case_dir_with_subdir, "partition/dir"),
        (_case_unique_id_with_dir_and_file_name, "partition/dir"),
        (_case_unique_id_with_file_name, "partition"),
    ],
)
def test_supabase_file_url_key_prefix_parent(
    file_url: SupabaseFileUrl, expected_key_prefix_parent: str
):
    assert file_url.key_prefix_parent == expected_key_prefix_parent


@pytest.mark.parametrize(
    "file_url",
    [
        (_case_dir_only),
        (_case_dir_with_subdir),
    ],
)
def test_supabase_file_url_is_directory(file_url: FileUrl):
    assert file_url.is_valid is True
    assert file_url.is_directory is True
    assert file_url.is_file is False


@pytest.mark.parametrize(
    "file_url",
    [
        (_case_unique_id_with_dir_and_file_name),
        (_case_unique_id_with_file_name),
    ],
)
def test_supabase_file_url_is_file(file_url: FileUrl):
    assert file_url.is_valid is True
    assert file_url.is_directory is False
    assert file_url.is_file is True


@pytest.mark.parametrize(
    "file_url, join_dir, join_file_name, expected_joined_path",
    [
        (_case_dir_only, None, None, "sb://fs_supabase_test/partition/dir"),
        (_case_dir_only, "join_dir", None, "sb://fs_supabase_test/partition/dir/join_dir"),
        (_case_dir_only, None, "join_f.txt", "sb://fs_supabase_test/partition/dir/join_f.txt"),
        (
            _case_dir_only,
            "join_dir",
            "join_f.txt",
            "sb://fs_supabase_test/partition/dir/join_dir/join_f.txt",
        ),
    ],
)
def test_supabase_file_url_join(
    file_url: SupabaseFileUrl,
    join_dir: str | None,
    join_file_name: str | None,
    expected_joined_path: str,
):
    joined_file_url = file_url.join(directory=join_dir, file_name=join_file_name)
    assert joined_file_url.as_str_internal() == expected_joined_path


@pytest.mark.parametrize(
    "file_url, expected_parent_path",
    [
        (_case_dir_only, "sb://fs_supabase_test/partition"),
        (_case_dir_with_subdir, "sb://fs_supabase_test/partition/dir"),
        (_case_unique_id_with_dir_and_file_name, "sb://fs_supabase_test/partition/dir"),
        (_case_unique_id_with_file_name, "sb://fs_supabase_test/partition"),
        (
            _case_unique_id_with_dir_and_file_name_encrypted,
            "sb://fs_supabase_test/partition/dir",
        ),
    ],
)
def test_supabase_file_url_parent(file_url: SupabaseFileUrl, expected_parent_path: str):
    parent_file_url = file_url.parent()
    assert parent_file_url.as_str_internal() == expected_parent_path



import asyncio
import datetime
import uuid
from typing import Any, get_args

import gotrue
import gotrue.errors
import gotrue.helpers
import psycopg
from argon2 import PasswordHasher
from fastapi_lifespan_manager import LifespanManager
from pydantic import Field
from supabase.client import AsyncClient as NativeSupabaseAsyncClient
from supabase.client import AsyncSupabaseAuthClient as NativeSupabaseAuthAsyncClient

# AsyncAuthClient as NativeSupabaseAuthAsyncClient,
from supabase.client import (
    create_async_client,
)
from supabase.lib.client_options import AsyncClientOptions as NativeSupabaseAsyncClientOptions

from apppy.auth.errors.user import (
    UserSessionRefreshMissingSessionError,
    UserSignInInvalidCredentialsError,
    UserSignInServerError,
    UserSignOutServerError,
    UserSignUpInvalidCredentialsError,
    UserSignUpServerError,
    UserSignUpTooManyRetriesError,
)
from apppy.db.postgres import PostgresClient
from apppy.env import Env, EnvSettings
from apppy.logger import WithLogger


class SupabaseAuthSessionStorage(gotrue.AsyncSupportedStorage, WithLogger):
    """
    A storage implementation for Supabase Auth that uses the FastAPI session storage.

    This allows us to persist state across requests, which is required for OAuth flows.
    """

    def __init__(self, session):
        self._session = session

    async def get_item(self, key: str) -> str | None:
        return self._session.get(key)

    async def set_item(self, key: str, value: str) -> None:
        self._session[key] = value

    async def remove_item(self, key: str) -> None:
        self._session.pop(key, None)


class SupabaseAuthSettings(EnvSettings):
    # SUPABASE_AUTH_API_ANON_KEY
    api_anon_key: str = Field()
    # SUPABASE_AUTH_API_KEY
    api_key: str = Field(exclude=True)
    # SUPABASE_AUTH_API_URL
    api_url: str = Field()
    # SUPABASE_AUTH_MAX_USER_SIGNUP_RETRIES
    max_user_signup_retries: int = Field(default=2)
    # SUPABASE_AUTH_REQUIRE_EMAIL_CONFIRMATION
    require_email_confirmation: bool = Field(default=True)

    def __init__(self, env: Env) -> None:
        super().__init__(env=env, domain_prefix="SUPABASE_AUTH")


class SupabaseAuthAdmin(WithLogger):
    def __init__(
        self,
        settings: SupabaseAuthSettings,
        lifespan: LifespanManager,
        postgres_client: PostgresClient,
    ):
        self._settings = settings
        self._postgres_client = postgres_client

        lifespan.add(self._auth_admin_async_client)

    async def _auth_admin_async_client(self):
        native_anon_async_client = await create_async_client(
            supabase_url=self._settings.api_url,
            supabase_key=self._settings.api_key,
        )

        self._supabase_auth_admin = native_anon_async_client.auth.admin
        yield {"supbase_auth_admin": self._supabase_auth_admin}

        self._logger.info("Closing Supabase Auth Admin client")
        await self._supabase_auth_admin.close()

    def _normalize_result_row(self, row: dict) -> dict:
        def _normalize_key(key: str):
            if key == "raw_app_meta_data":
                return "app_metadata"
            elif key == "raw_user_meta_data":
                return "user_metadata"

            return key

        return {
            _normalize_key(k): str(v) if isinstance(v, uuid.UUID | datetime.datetime) else v
            for k, v in row.items()
        }

    async def create_user(self, attributes: gotrue.AdminUserAttributes) -> gotrue.User:
        auth_user_query = """
            SELECT *
            FROM auth.users
            WHERE email = %(email)s
            LIMIT 1
        """
        try:
            result_set = await self._postgres_client.db_query_async(
                auth_user_query, {"email": attributes["email"]}
            )

            if result_set is None or len(result_set) != 1:
                user_resp = await self._supabase_auth_admin.create_user(attributes)
                return user_resp.user
            else:
                user_dict = self._normalize_result_row(result_set[0])
                user_resp = gotrue.helpers.parse_user_response(user_dict)
                return user_resp.user
        except psycopg.DatabaseError as e:
            self._logger.exception("Lookup during create user encountered an api error")
            raise UserSignUpServerError("auth_user_lookup_error") from e
        except gotrue.errors.AuthApiError as e:
            self._logger.exception(
                "Create user encountered an auth api error",
                extra={"error_name": e.name},
            )
            raise UserSignUpServerError(e.code) from e


class SupabaseAuth(WithLogger):
    def __init__(
        self,
        settings: SupabaseAuthSettings,
        supabase_auth_admin: SupabaseAuthAdmin,
    ) -> None:
        self._settings = settings
        self._password_hasher = PasswordHasher()
        self._supabase_auth_admin = supabase_auth_admin

    async def auth_async_client(
        self, session: dict[str, Any] | None = None
    ) -> NativeSupabaseAuthAsyncClient:
        if session is not None:
            client_options = NativeSupabaseAsyncClientOptions(
                storage=SupabaseAuthSessionStorage(session)
            )
        else:
            client_options = NativeSupabaseAsyncClientOptions()

        native_anon_async_client: NativeSupabaseAsyncClient = await create_async_client(
            supabase_url=self._settings.api_url,
            supabase_key=self._settings.api_anon_key,
            options=client_options,
        )

        return native_anon_async_client.auth

    def is_native_supabase_provider(self, provider: str) -> bool:
        return provider in get_args(gotrue.Provider)

    @property
    def require_email_confirmation(self) -> bool:
        return self._settings.require_email_confirmation

    async def user_session_login_id_token(
        self,
        credentials: gotrue.SignInWithIdTokenCredentials,
    ) -> gotrue.AuthResponse:
        try:
            auth_async_client: NativeSupabaseAuthAsyncClient = await self.auth_async_client()
            auth_resp: gotrue.AuthResponse = await auth_async_client.sign_in_with_id_token(
                credentials
            )

            return auth_resp
        except gotrue.errors.AuthApiError as e:
            self._logger.exception(
                "Login with id token encountered an auth api error",
                extra={"error_name": e.name, "error_code": e.code, "error_message": e.message},
            )
            raise UserSignInServerError(e.code) from e
        except gotrue.errors.CustomAuthError as e:
            self._logger.exception(
                "Login with id token encountered an auth custom error",
                extra={"error_name": e.name, "error_code": e.code, "error_message": e.message},
            )
            raise UserSignInServerError(e.code) from e

    async def user_session_login_oauth_finalize(
        self,
        session: dict[str, Any],
        params: gotrue.CodeExchangeParams,
    ) -> gotrue.AuthResponse:
        try:
            auth_async_client: NativeSupabaseAuthAsyncClient = await self.auth_async_client(session)
            auth_resp: gotrue.AuthResponse = await auth_async_client.exchange_code_for_session(
                params
            )

            return auth_resp
        except gotrue.errors.AuthApiError as e:
            self._logger.exception(
                "Oauth login finalize encountered an auth api error",
                extra={"error_name": e.name, "error_code": e.code, "error_message": e.message},
            )
            raise UserSignInServerError(e.code) from e
        except gotrue.errors.CustomAuthError as e:
            self._logger.exception(
                "OAuth login finalize encountered an auth custom error",
                extra={"error_name": e.name, "error_code": e.code, "error_message": e.message},
            )
            raise UserSignInServerError(e.code) from e

    async def user_session_login_oauth(
        self,
        session: dict[str, Any],
        credentials: gotrue.SignInWithOAuthCredentials,
    ) -> gotrue.OAuthResponse:
        try:
            auth_async_client: NativeSupabaseAuthAsyncClient = await self.auth_async_client(session)
            oauth_resp: gotrue.OAuthResponse = await auth_async_client.sign_in_with_oauth(
                credentials
            )

            return oauth_resp
        except gotrue.errors.AuthApiError as e:
            self._logger.exception(
                "Login with oauth encountered an auth api error",
                extra={"error_name": e.name, "error_code": e.code, "error_message": e.message},
            )
            raise UserSignInServerError(e.code) from e
        except gotrue.errors.CustomAuthError as e:
            self._logger.exception(
                "Login with oauth encountered an auth custom error",
                extra={"error_name": e.name, "error_code": e.code, "error_message": e.message},
            )
            raise UserSignInServerError(e.code) from e

    async def user_session_login_password(
        self, credentials: gotrue.SignInWithPasswordCredentials
    ) -> gotrue.AuthResponse:
        try:
            auth_async_client: NativeSupabaseAuthAsyncClient = await self.auth_async_client()
            auth_resp: gotrue.AuthResponse = await auth_async_client.sign_in_with_password(
                credentials
            )

            return auth_resp
        except gotrue.errors.AuthApiError as e:
            if e.message == "Invalid login credentials":
                raise UserSignInInvalidCredentialsError() from e

            self._logger.exception(
                "Login with password encountered an auth api error",
                extra={"error_name": e.name, "error_code": e.code, "error_message": e.message},
            )
            raise UserSignInServerError(e.code) from e
        except gotrue.errors.CustomAuthError as e:
            self._logger.exception(
                "Login with password encountered an auth custom error",
                extra={"error_name": e.name, "error_code": e.code, "error_message": e.message},
            )
            raise UserSignInServerError(e.code) from e

    async def user_session_logout(self, options: gotrue.SignOutOptions) -> None:
        auth_async_client: NativeSupabaseAuthAsyncClient = await self.auth_async_client()
        try:
            await auth_async_client.sign_out(options)
        except gotrue.errors.AuthError as e:
            self._logger.exception(
                "Logout encountered an auth error",
                extra={"error_name": e.name, "error_code": e.code, "error_message": e.message},
            )
            raise UserSignOutServerError(e.code) from e

    async def user_session_refresh(self, refresh_token: str) -> gotrue.AuthResponse:
        try:
            auth_async_client: NativeSupabaseAuthAsyncClient = await self.auth_async_client()
            auth_resp: gotrue.AuthResponse = await auth_async_client.refresh_session(refresh_token)
            return auth_resp
        except gotrue.errors.AuthSessionMissingError as e:
            raise UserSessionRefreshMissingSessionError() from e

    async def user_sign_up(
        self, credentials: gotrue.SignUpWithPasswordCredentials
    ) -> gotrue.AuthResponse:
        attempt = 0
        while attempt < self._settings.max_user_signup_retries:
            try:
                auth_async_client = await self.auth_async_client()
                auth_resp: gotrue.AuthResponse = await auth_async_client.sign_up(credentials)
                return auth_resp
            except gotrue.errors.AuthRetryableError:
                self._logger.warning(
                    "Encountered an AuthRetryableError with signing up a Supabase user. Retrying.",
                    extra={
                        "attempt": attempt,
                        "email": credentials.get("email"),
                        "phone": credentials.get("phone"),
                    },
                )
                attempt += 1
                await asyncio.sleep(0.5)
            except gotrue.errors.AuthInvalidCredentialsError as e:
                self._logger.exception(
                    e.message,
                    extra={"email": credentials.get("email"), "phone": credentials.get("phone")},
                )
                raise UserSignUpInvalidCredentialsError() from e
            except gotrue.errors.AuthApiError as e:
                self._logger.exception(
                    "Sign up encountered an auth api error",
                    extra={"email": credentials.get("email"), "phone": credentials.get("phone")},
                )
                raise UserSignUpServerError(e.code) from e

        self._logger.error(
            "Too many retry attempts to sign up a Supabase user",
            extra={
                "attempts": attempt,
                "email": credentials.get("email"),
                "phone": credentials.get("phone"),
            },
        )
        raise UserSignUpTooManyRetriesError()

    async def user_sign_up_id_token(
        self,
        provider: str,
        attributes: gotrue.AdminUserAttributes,
    ) -> None:
        user: gotrue.User = await self._supabase_auth_admin.create_user(attributes)
        if user is None:
            raise UserSignUpServerError("user_creation_failed_in_sign_up_id_token")



from dataclasses import dataclass

import psycopg

# AsyncAuthClient as NativeSupabaseAuthAsyncClient,
from apppy.db.postgres import PostgresClient
from apppy.logger import WithLogger
from apppy.sb.identity.errors import MissingUserIdentityError, UserIdentityServerError


@dataclass
class UserIdentityResponse:
    name: str
    email: str
    phone: str


class SupabaseIdentity(WithLogger):
    def __init__(
        self,
        postgres_client: PostgresClient,
    ) -> None:
        self._postgres_client = postgres_client

    async def user_identity(self, user_id: str):
        auth_user_identity_query = """
            SELECT *
            FROM auth.identities
            WHERE user_id = %(user_id)s
            LIMIT 1
        """
        try:
            result_set = await self._postgres_client.db_query_async(
                auth_user_identity_query, {"user_id": user_id}
            )

            if result_set is not None and len(result_set) == 1:
                return result_set[0]

            raise MissingUserIdentityError()
        except psycopg.DatabaseError as e:
            self._logger.exception("Lookup for user identity encountered an api error")
            raise UserIdentityServerError("auth_user_identity_lookup_error") from e



from fastapi_lifespan_manager import LifespanManager
from pydantic import Field
from supabase import AsyncClient as NativeSupabaseAsyncClient
from supabase import Client as NativeSupabaseClient
from supabase.client import create_async_client, create_client

from apppy.env import Env, EnvSettings
from apppy.logger import WithLogger

_SUPABASE_MICRO_MAX_CONNS = 200
_SUPABASE_SMALL_MAX_CONNS = 400


class SupabaseClientSettings(EnvSettings):
    # SUPABASE_API_ANON_KEY
    api_anon_key: str = Field()
    # SUPABASE_API_KEY
    api_key: str = Field(exclude=True)
    # SUPABASE_API_URL
    api_url: str = Field()

    def __init__(self, env: Env) -> None:
        super().__init__(env=env, domain_prefix="SUPABASE")


class SupabaseClient(WithLogger):
    def __init__(
        self,
        settings: SupabaseClientSettings,
        lifespan: LifespanManager | None,
    ) -> None:
        self._settings = settings
        self._native_internal_client: NativeSupabaseClient = create_client(
            supabase_url=settings.api_url, supabase_key=settings.api_key
        )

        if lifespan is not None:
            lifespan.add(self.__create_async_client)

    async def __create_async_client(self):
        self._logger.info("Creating native_internal_async_client")
        self._native_internal_async_client = await create_async_client(
            supabase_url=self._settings.api_url,
            supabase_key=self._settings.api_key,
        )

        yield {"native_internal_async_client": self._native_internal_async_client}

        self._logger.info("Closing native_internal_async_client")

    @property
    def internal_client(self) -> NativeSupabaseClient:
        return self._native_internal_client

    @property
    def internal_async_client(self) -> NativeSupabaseAsyncClient:
        assert self._native_internal_async_client is not None, (
            "Attempting to use internal_async_client when a LifespanManager was never supplied "
            + "(try supplying a LifespanManager or using internal_client)"
        )
        return self._native_internal_async_client



import pytest_asyncio
from fastapi_lifespan_manager import LifespanManager

from apppy.env import Env
from apppy.env.fixtures import current_test_name
from apppy.fs import FileSystem, FileSystemSettings, FileUrl
from apppy.sb.fs import SupabaseFileSystem, SupabaseFileSystemSettings


@pytest_asyncio.fixture(scope="session")
async def sb_fs():
    fs_env: Env = Env.load(name=current_test_name())
    fs_settings = FileSystemSettings(fs_env)
    fs = FileSystem(fs_settings)

    fs_sb_settings = SupabaseFileSystemSettings(fs_env)
    lifespan = LifespanManager()
    fs_sb = SupabaseFileSystem(fs_env, fs_sb_settings, lifespan, fs)

    dummy_app = object()
    async with lifespan(dummy_app) as state:
        _ = fs_sb, state
        yield fs


@pytest_asyncio.fixture(scope="session")
async def sb_fs_encrypted():
    fs_env: Env = Env.load(
        name=current_test_name(),
        overrides={
            "encrypt_passphrase": "my-sb-fs-passphrase",
            "encrypt_salt": "some-salt",
        },
    )
    fs_settings = FileSystemSettings(fs_env)
    fs = FileSystem(fs_settings)

    fs_sb_settings = SupabaseFileSystemSettings(fs_env)
    lifespan = LifespanManager()
    fs_sb = SupabaseFileSystem(fs_env, fs_sb_settings, lifespan, fs)

    dummy_app = object()
    async with lifespan(dummy_app) as state:
        _ = fs_sb, state
        yield fs


@pytest_asyncio.fixture
async def sb_file_write(sb_fs: FileSystem):
    async def _file_url_from_file_write(partition: str, directory: str) -> FileUrl:
        file_url: FileUrl = sb_fs.new_file_url_external(
            protocol="sb",
            external_id=None,
            partition=partition,
            directory=directory,
            file_name="file.txt",
        )

        file_url, _ = await sb_fs.write_text(
            file_url, "The quick brown fox jumped over the lazy dogs."
        )
        return file_url

    return _file_url_from_file_write


@pytest_asyncio.fixture
async def sb_file_write_encrypted(sb_fs_encrypted: FileSystem):
    async def _file_url_from_file_write(partition: str, directory: str) -> FileUrl:
        file_url: FileUrl = sb_fs_encrypted.new_file_url_external(
            protocol="sb",
            external_id=None,
            partition=partition,
            directory=directory,
            file_name="file.txt",
        )

        file_url, _ = await sb_fs_encrypted.write_text(
            file_url, "The quick brown fox jumped over the lazy dogs."
        )
        return file_url

    return _file_url_from_file_write
